/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License;
 * you may not use this file except in compliance with the Elastic License.
 */
package org.elasticsearch.integration.ldap;

import org.elasticsearch.ElasticsearchSecurityException;
import org.elasticsearch.action.DocWriteResponse;
import org.elasticsearch.action.ActionFuture;
import org.elasticsearch.action.get.GetResponse;
import org.elasticsearch.action.index.IndexResponse;
import org.elasticsearch.client.Client;
import org.elasticsearch.common.Nullable;
import org.elasticsearch.common.bytes.BytesArray;
import org.elasticsearch.common.logging.ESLoggerFactory;
import org.elasticsearch.common.settings.SecureString;
import org.elasticsearch.common.settings.Settings;
import org.elasticsearch.common.xcontent.XContentType;
import org.elasticsearch.test.ESTestCase;
import org.elasticsearch.test.SecurityIntegTestCase;
import org.elasticsearch.xpack.security.action.rolemapping.PutRoleMappingRequestBuilder;
import org.elasticsearch.xpack.security.action.rolemapping.PutRoleMappingResponse;
import org.elasticsearch.xpack.security.authc.ldap.LdapRealm;
import org.elasticsearch.xpack.security.authc.support.UsernamePasswordToken;
import org.elasticsearch.xpack.security.client.SecurityClient;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static org.elasticsearch.common.xcontent.XContentFactory.jsonBuilder;
import static org.elasticsearch.xpack.security.authc.ldap.support.LdapSearchScope.ONE_LEVEL;
import static org.elasticsearch.xpack.security.authc.ldap.support.LdapSearchScope.SUB_TREE;
import static org.elasticsearch.xpack.security.authc.support.UsernamePasswordToken.BASIC_AUTH_HEADER;
import static org.elasticsearch.xpack.security.test.SecurityTestUtils.writeFile;
import static org.hamcrest.Matchers.equalTo;

/**
 * This test assumes all subclass tests will be of type SUITE.  It picks a random realm configuration for the tests, and
 * writes a group to role mapping file for each node.
 */
public abstract class AbstractAdLdapRealmTestCase extends SecurityIntegTestCase {

    public static final String XPACK_SECURITY_AUTHC_REALMS_EXTERNAL = "xpack.security.authc.realms.external";
    public static final String PASSWORD = "NickFuryHeartsES";
    public static final String ASGARDIAN_INDEX = "gods";
    public static final String PHILANTHROPISTS_INDEX = "philanthropists";
    public static final String SECURITY_INDEX = "security";

    private static final RoleMappingEntry[] AD_ROLE_MAPPING = new RoleMappingEntry[] {
            new RoleMappingEntry(
                    "SHIELD:  [ \"CN=SHIELD,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" ]",
                    "{ \"roles\":[\"SHIELD\"], \"enabled\":true, \"rules\":" +
                            "{\"field\": {\"groups\": \"CN=SHIELD,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\"} } }"
            ),
            new RoleMappingEntry(
                    "Avengers:  [ \"CN=Avengers,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" ]",
                    "{ \"roles\":[\"Avengers\"], \"enabled\":true, \"rules\":" +
                            "{ \"field\": { \"groups\" : \"CN=Avengers,CN=Users,*\" } } }"
            ),
            new RoleMappingEntry(
                    "Gods:  [ \"CN=Gods,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" ]",
                    "{ \"roles\":[\"Gods\"], \"enabled\":true, \"rules\":{\"any\": [" +
                            " { \"field\":{ \"groups\":    \"CN=Gods,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" } }," +
                            " { \"field\":{ \"groups\": \"CN=Deities,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" } } " +
                            "] } }"
            ),
            new RoleMappingEntry(
                    "Philanthropists:  [ \"CN=Philanthropists,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" ]",
                    "{ \"roles\":[\"Philanthropists\"], \"enabled\":true, \"rules\": { \"all\": [" +
                            " { \"field\": { \"groups\" : \"CN=Philanthropists,CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com\" } }," +
                            " { \"field\": { \"realm.name\" : \"external\" } } " +
                            "] } }"
            )
    };

    private static final RoleMappingEntry[] OLDAP_ROLE_MAPPING = new RoleMappingEntry[] {
            new RoleMappingEntry(
                    "SHIELD: [ \"cn=SHIELD,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" ]",
                    "{ \"roles\": [\"SHIELD\"], \"enabled\":true, \"rules\":" +
                            " {\"field\": {\"groups\": \"cn=SHIELD,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" } } }"
            ),
            new RoleMappingEntry(
                    "Avengers: [ \"cn=Avengers,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" ]",
                    "{ \"roles\": [\"Avengers\"], \"enabled\":true, \"rules\":" +
                            " {\"field\": {\"groups\": \"cn=Avengers,ou=people,*\" } } }"
            ),
            new RoleMappingEntry(
                    "Gods: [ \"cn=Gods,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" ]",
                    "{ \"roles\" : [ \"Gods\" ], \"enabled\":true, \"rules\" : { \"any\": [" +
                            " {\"field\": {\"groups\":    \"cn=Gods,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" } }," +
                            " {\"field\": {\"groups\": \"cn=Deities,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" } } " +
                            "] } }"
            ),
            new RoleMappingEntry(
                    "Philanthropists: [ \"cn=Philanthropists,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" ]",
                    "{ \"roles\" : [ \"Philanthropists\" ], \"enabled\":true, \"rules\" : { \"all\": [" +
                            " { \"field\": { \"groups\" : \"cn=Philanthropists,ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com\" } }, " +
                            " { \"field\": { \"realm.name\" : \"external\" } } " +
                            "] } }"
            )
    };

    protected static final String TESTNODE_KEYSTORE = "/org/elasticsearch/xpack/security/transport/ssl/certs/simple/testnode.jks";
    protected static RealmConfig realmConfig;
    protected static List<RoleMappingEntry> roleMappings;
    protected static boolean useGlobalSSL;

    @BeforeClass
    public static void setupRealm() {
        realmConfig = randomFrom(RealmConfig.values());
        roleMappings = realmConfig.selectRoleMappings(ESTestCase::randomBoolean);
        useGlobalSSL = randomBoolean();
        ESLoggerFactory.getLogger("test").info("running test with realm configuration [{}], with direct group to role mapping [{}]. " +
                "Settings [{}]", realmConfig, realmConfig.mapGroupsAsRoles, realmConfig.settings.getAsMap());
    }

    @AfterClass
    public static void cleanupRealm() {
        realmConfig = null;
    }

    @Override
    protected Settings nodeSettings(int nodeOrdinal) {
        final RealmConfig realm = AbstractAdLdapRealmTestCase.realmConfig;
        Path store = getDataPath(TESTNODE_KEYSTORE);
        Settings.Builder builder = Settings.builder();
        if (useGlobalSSL) {
            builder.put(super.nodeSettings(nodeOrdinal).filter((s) -> s.startsWith("xpack.ssl.") == false))
                    .put(sslSettingsForStore(store, "testnode"));
        } else {
            builder.put(super.nodeSettings(nodeOrdinal));
        }
        builder.put(buildRealmSettings(realm, store));
        return builder.build();
    }

    protected Settings buildRealmSettings(RealmConfig realm, Path store) {
        Settings.Builder builder = Settings.builder();
        builder.put(realm.buildSettings(store, "testnode"));
        configureRoleMappings(builder);
        return builder.build();
    }

    @Before
    public void setupRoleMappings() throws Exception {
        assertSecurityIndexActive();

        List<String> content = getRoleMappingContent(RoleMappingEntry::getNativeContent);
        if (content.isEmpty()) {
            return;
        }
        SecurityClient securityClient = securityClient();
        Map<String, ActionFuture<PutRoleMappingResponse>> futures
                = new LinkedHashMap<>(content.size());
        for (int i = 0; i < content.size(); i++) {
            final String name = "external_" + i;
            final PutRoleMappingRequestBuilder builder = securityClient.preparePutRoleMapping(
                    name, new BytesArray(content.get(i)), XContentType.JSON);
            futures.put(name, builder.execute());
        }
        for (String mappingName : futures.keySet()) {
            final PutRoleMappingResponse response = futures.get(mappingName).get();
            logger.info("Created native role-mapping {} : {}", mappingName, response.isCreated());
        }
    }

    @After
    public void cleanupSecurityIndex() throws Exception {
        super.deleteSecurityIndex();
    }

    private List<String> getRoleMappingContent(Function<RoleMappingEntry, String> contentFunction) {
        return roleMappings.stream()
                .map(contentFunction)
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
    }

    @Override
    protected Settings transportClientSettings() {
        if (useGlobalSSL) {
            Path store = getDataPath(TESTNODE_KEYSTORE);
            return Settings.builder()
                    .put(super.transportClientSettings().filter((s) -> s.startsWith("xpack.ssl.") == false))
                    .put(sslSettingsForStore(store, "testnode"))
                    .build();
        } else {
            return super.transportClientSettings();
        }
    }

    @Override
    protected boolean useGeneratedSSLConfig() {
        return useGlobalSSL == false;
    }

    protected final void configureRoleMappings(Settings.Builder builder) {
        String content = getRoleMappingContent(RoleMappingEntry::getFileContent).stream().collect(Collectors.joining("\n"));
        Path nodeFiles = createTempDir();
        String file = writeFile(nodeFiles, "role_mapping.yml", content);
        builder.put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".files.role_mapping", file);
    }

    @Override
    protected String configRoles() {
        return super.configRoles() +
                "\n" +
                "Avengers:\n" +
                "  cluster: [ NONE ]\n" +
                "  indices:\n" +
                "    - names: 'avengers'\n" +
                "      privileges: [ all ]\n" +
                "SHIELD:\n" +
                "  cluster: [ NONE ]\n" +
                "  indices:\n" +
                "    - names: '" + SECURITY_INDEX + "'\n" +
                "      privileges: [ all ]\n" +
                "Gods:\n" +
                "  cluster: [ NONE ]\n" +
                "  indices:\n" +
                "    - names: '" + ASGARDIAN_INDEX + "'\n" +
                "      privileges: [ all ]\n" +
                "Philanthropists:\n" +
                "  cluster: [ NONE ]\n" +
                "  indices:\n" +
                "    - names: '" + PHILANTHROPISTS_INDEX + "'\n" +
                "      privileges: [ all ]\n";
    }

    protected void assertAccessAllowed(String user, String index) throws IOException {
        Client client = client().filterWithHeader(Collections.singletonMap(BASIC_AUTH_HEADER, userHeader(user, PASSWORD)));
        IndexResponse indexResponse = client.prepareIndex(index, "type").
                setSource(jsonBuilder()
                        .startObject()
                        .field("name", "value")
                        .endObject())
                .execute().actionGet();

        assertEquals("user " + user + " should have write access to index " + index,
                DocWriteResponse.Result.CREATED, indexResponse.getResult());

        refresh();

        GetResponse getResponse = client.prepareGet(index, "type", indexResponse.getId())
                .get();

        assertThat("user " + user + " should have read access to index " + index, getResponse.getId(), equalTo(indexResponse.getId()));
    }

    protected void assertAccessDenied(String user, String index) throws IOException {
        try {
            client().filterWithHeader(Collections.singletonMap(BASIC_AUTH_HEADER, userHeader(user, PASSWORD)))
                    .prepareIndex(index, "type").
                    setSource(jsonBuilder()
                            .startObject()
                            .field("name", "value")
                            .endObject())
                    .execute().actionGet();
            fail("Write access to index " + index + " should not be allowed for user " + user);
        } catch (ElasticsearchSecurityException e) {
            // expected
        }
        refresh();
    }

    protected static String userHeader(String username, String password) {
        return UsernamePasswordToken.basicAuthHeaderValue(username, new SecureString(password.toCharArray()));
    }

    private Settings sslSettingsForStore(Path store, String password) {
        return Settings.builder()
                .put("xpack.ssl.keystore.path", store)
                .put("xpack.ssl.keystore.password", password)
                .put("xpack.ssl.verification_mode", "certificate")
                .put("xpack.ssl.truststore.path", store)
                .put("xpack.ssl.truststore.password", password).build();
    }

    static class RoleMappingEntry {
        @Nullable
        public final String fileContent;
        @Nullable
        public final String nativeContent;

        RoleMappingEntry(@Nullable String fileContent, @Nullable String nativeContent) {
            this.fileContent = fileContent;
            this.nativeContent = nativeContent;
        }

        String getFileContent() {
            return fileContent;
        }

        String getNativeContent() {
            return nativeContent;
        }

        RoleMappingEntry pickEntry(Supplier<Boolean> shouldPickFileContent) {
            if (nativeContent == null) {
                return new RoleMappingEntry(fileContent, null);
            }
            if (fileContent == null) {
                return new RoleMappingEntry(null, nativeContent);
            }
            if (shouldPickFileContent.get()) {
                return new RoleMappingEntry(fileContent, null);
            } else {
                return new RoleMappingEntry(null, nativeContent);
            }
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) {
                return true;
            }
            if (o == null || getClass() != o.getClass()) {
                return false;
            }

            final RoleMappingEntry that = (RoleMappingEntry) o;
            return Objects.equals(this.fileContent, that.fileContent)
                    && Objects.equals(this.nativeContent, that.nativeContent);
        }

        @Override
        public int hashCode() {
            int result = Objects.hashCode(fileContent);
            result = 31 * result + Objects.hashCode(nativeContent);
            return result;
        }
    }

    /**
     * Represents multiple possible configurations for active directory and ldap
     */
    enum RealmConfig {

        AD(false, AD_ROLE_MAPPING,
                Settings.builder()
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".type", LdapRealm.AD_TYPE)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".domain_name", "ad.test.elasticsearch.com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL
                                + ".group_search.base_dn", "CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".group_search.scope", randomBoolean() ? SUB_TREE : ONE_LEVEL)
                        .build()),

        AD_SSL(false, AD_ROLE_MAPPING,
                Settings.builder()
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".type", LdapRealm.AD_TYPE)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".domain_name", "ad.test.elasticsearch.com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL
                                + ".group_search.base_dn", "CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".group_search.scope", randomBoolean() ? SUB_TREE : ONE_LEVEL)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".url", "ldap://ad.test.elasticsearch.com:389")
                        .build()),

        AD_LDAP_GROUPS_FROM_SEARCH(true, AD_ROLE_MAPPING,
                Settings.builder()
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".type", LdapRealm.LDAP_TYPE)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".url", "ldaps://ad.test.elasticsearch.com:636")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL
                                + ".group_search.base_dn", "CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".group_search.scope", randomBoolean() ? SUB_TREE : ONE_LEVEL)
                        .putArray(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".user_dn_templates",
                                "cn={0},CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com")
                        .build()),

        AD_LDAP_GROUPS_FROM_ATTRIBUTE(true, AD_ROLE_MAPPING,
                Settings.builder()
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".type", LdapRealm.LDAP_TYPE)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".url", "ldaps://ad.test.elasticsearch.com:636")
                        .putArray(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".user_dn_templates",
                                "cn={0},CN=Users,DC=ad,DC=test,DC=elasticsearch,DC=com")
                        .build()),

        OLDAP(false, OLDAP_ROLE_MAPPING,
                Settings.builder()
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".type", LdapRealm.LDAP_TYPE)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".url", "ldaps://54.200.235.244:636")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".group_search.base_dn",
                                "ou=people, dc=oldap, dc=test, dc=elasticsearch, dc=com")
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".group_search.scope", randomBoolean() ? SUB_TREE : ONE_LEVEL)
                        .putArray(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".user_dn_templates",
                                "uid={0},ou=people,dc=oldap,dc=test,dc=elasticsearch,dc=com")
                        .build());

        final boolean mapGroupsAsRoles;
        final boolean loginWithCommonName;
        private final RoleMappingEntry[] roleMappings;
        final Settings settings;

        RealmConfig(boolean loginWithCommonName, RoleMappingEntry[] roleMappings, Settings settings) {
            this.settings = settings;
            this.loginWithCommonName = loginWithCommonName;
            this.roleMappings = roleMappings;
            this.mapGroupsAsRoles = randomBoolean();
        }

        public Settings buildSettings(Path store, String password) {
            return buildSettings(store, password, 1);
        }

        protected Settings buildSettings(Path store, String password, int order) {
            Settings.Builder builder = Settings.builder()
                    .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".order", order)
                    .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".hostname_verification", false)
                    .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".unmapped_groups_as_roles", mapGroupsAsRoles)
                    .put(this.settings);
            if (useGlobalSSL == false) {
                builder.put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".ssl.truststore.path", store)
                        .put(XPACK_SECURITY_AUTHC_REALMS_EXTERNAL + ".ssl.truststore.password", password);
            }

            return builder.build();
        }

        public List<RoleMappingEntry> selectRoleMappings(Supplier<Boolean> shouldPickFileContent) {
            // if mapGroupsAsRoles is turned on we use empty role mapping
            if (mapGroupsAsRoles) {
                return Collections.emptyList();
            } else {
                return Arrays.stream(this.roleMappings)
                        .map(e -> e.pickEntry(shouldPickFileContent))
                        .collect(Collectors.toList());
            }
        }
    }
}
